<?php

/**
 *  Classe construída com o proposito de organizar rotinas uteis para facilitar a manipulacao de entidades Doctrine, a partir de arrays de informacões.
 *  Possibilita as operações de criação, alteração e remoção de dados a partir de arrays com as informações das entidades. Permite que a ação de 'persist' no momento da criação
 *  ou atualização seja postergado para um momento oportuno, pensando em situações em que há a manipulação de entidades mestre-histórico. A rotina 'persistPending' realiza o 'persist'
 *  dessas entidades.
 * 
 *  Principais rotinas:
 *      - setNewEntity;
 *      - updateEntity;
 *      - removeEntity;
 * 
 *  Implmenta definições de EntityPersistInstance.
 *  
 */

namespace Ems\Util;

use Doctrine\ORM\EntityManagerInterface;

class DoctrineEntityUtil implements EntityPersistInstance {

    const SET_DEFAULT_FIELDS    = 'SET_DEFAULT_FIELDS';
    const PERSIST               = 'PERSIST';
    const FLUSH                 = 'FLUSH';
    const PENDING               = 'PENDING';

    const DEFAULT_STORE_ENTITY_CONFIG = [
        'PERSIST'            => true,
        'FLUSH'              => true,
        'PENDING'            => false,
    ];

    const DEFAULT_UPDATE_CONFIG = [
        'SET_DEFAULT_FIELDS' => true,
        'PERSIST'            => true,
        'FLUSH'              => true,
        'PENDING'            => false,
    ];

    const DEFAULT_INSERT_CONFIG = [
        'SET_DEFAULT_FIELDS' => true,
        'PERSIST'            => true,
        'FLUSH'              => true,
        'PENDING'            => false,
    ];

    const DEFAULT_REMOVE_CONFIG = [
        'FLUSH' =>  true,
    ];
    
    private $persistPendingEntities = [];

    /**
     * Cria uma nova entidade Doctrine a partir da row e nome do repositorio informados. A operacao 'persist' pode ser adiada atraves da definicao da propriedade 'PENDING'
     * como verdadeiro e acionar 'persistPending' no momento desejado.
     * 
     * @param  string  $entityRepositoryName
     * @param  array   $row
     * @param  array   $config = self::DEFAULT_INSERT_CONFIG (Nullable)
     * 
     * @return object
     */
    public function setNewEntity(
        string  $entityRepositoryName,
        array   $row,
        ?array  $config = self::DEFAULT_INSERT_CONFIG
    ): object {
        $config = array_merge(self::DEFAULT_INSERT_CONFIG, $config);

        $newEntity = new $entityRepositoryName;
        
        $finalRow = $row;
        if ($config[self::SET_DEFAULT_FIELDS]) {
            $finalRow = array_merge($this->getNewEntityDefaultFields(), $row);
        }

        $this->setEntityValuesByRow($newEntity, $finalRow);

        $this->storeEntity($newEntity, $config);
        
        return $newEntity;
    }

    /**
     * Atualiza as propriedades de uma entidade existente a partir de uma row. As chaves, informadas como entityKeysNames sao utilizadas para buscar a entidade e nao sao
     * alteradas, caso o registro exista. Caso o registro nao exista, um novo e criado.
     * 
     * @param   string  $entityRepositoryName
     * @param   array   $entityKeysNames
     * @param   array   $row
     * @param   array   $config = self::DEFAULT_UPDATE_CONFIG (Nullable)
     * 
     * @return object
     */
    public function updateEntity(
        string  $entityRepositoryName,
        array   $entityKeysNames,
        array   $row,
        ?array  $config = self::DEFAULT_UPDATE_CONFIG
    ): object {
        $config = array_merge(self::DEFAULT_UPDATE_CONFIG, $config);

        $existingEntity = $this->findEntityByRow($entityRepositoryName, $entityKeysNames, $row);
        if (empty($existingEntity)) {
            return $this->setNewEntity($entityRepositoryName, $row, $config);
        }

        $finalRow = $row;

        if ($config[self::SET_DEFAULT_FIELDS]) {
            $finalRow = array_merge($this->getUpdatedEntityDefaultFields(), $row);
        }

        $finalRow = $this->unsetEntityKeysOnRow($finalRow, $entityKeysNames);

        $this->setEntityValuesByRow($existingEntity, $finalRow);

        $this->storeEntity($existingEntity, $config);

        return $existingEntity;
    }

    /**
     * Remove uma entidade a partir de uma row informada. A entidade removida e retornada, caso o registro exista de fato. Caso contrario, falso e retornado.
     * 
     * @param   string  $entityRepositoryName
     * @param   array   $entityKeysNames
     * @param   array   $row
     * @param   array   $config = self::DEFAULT_REMOVE_CONFIG (Nullable)

     * @return object|false
     */
    public function removeEntity(
        string  $entityRepositoryName,
        array   $entityKeysNames,
        array   $row,
        ?array  $config = self::DEFAULT_REMOVE_CONFIG
    ) {
        $config = array_merge(self::DEFAULT_REMOVE_CONFIG, $config);

        $existingEntity = $this->findEntityByRow($entityRepositoryName, $entityKeysNames, $row);
        if (empty($existingEntity)) {
            return false;
        }

        $this->getEntityManager()->remove($existingEntity);

        if ($config[self::FLUSH]) {
            $this->getEntityManager()->flush();
        }
        
        return $existingEntity;
    }

    /**
     * Realiza a operacao de persist nas entidades separadas, no momento do insert ou update. O array de entidades pendentes e limpo apos a operacao.
     * 
     * @return void
     */
    public function persistPending(): void {
        foreach (array_reverse($this->persistPendingEntities) as $entity) {
            $this->getEntityManager()->persist($entity);
        }
        
        $this->persistPendingEntities = [];
    }

    /**
     * Retorna as entidades pendentes de persist.
     * 
     * @return array
     */
    public function getPersistPending(): array {
        return $this->persistPendingEntities;
    }

    private function setEntityValuesByRow(
        object  $entity,
        array   $row
    ): object {
        foreach ($row as $field => $value) {
            $setMethod = $this->buildSetEntityMethod($field);
            if (!method_exists($entity, $setMethod)) {
                continue;
            }
            
            $entity->$setMethod($value);
        }

        return $entity;
    }

    private function storeEntity(
        object  $entity,
        array   $config
    ): object {
        $config = array_merge(self::DEFAULT_STORE_ENTITY_CONFIG, $config);

        if ($config[self::PERSIST]) {
            $this->getEntityManager()->persist($entity);
        } else if ($config[self::PENDING]) {
            $this->persistPendingEntities[] = $entity;
        }
        
        if ($config[self::FLUSH]) {
            $this->getEntityManager()->flush($entity);
        }
        
        return $entity;
    }

    private function findEntityByRow(
        string  $entityRepositoryName,
        array   $entityKeysNames,
        array   $row
    ): ?object {
        $findFilters = $this->getEntityKeysFindFiltersInRow($entityKeysNames, $row);
        $searchedEntity = $this->getEntityManager()->getRepository($entityRepositoryName)->findOneBy($findFilters);
        return $searchedEntity;
    }

    private function getEntityKeysFindFiltersInRow(
        array   $entityKeysNames,
        array   $row
    ): array {
        $findData = [];

        foreach ($entityKeysNames as $entityKeyName) {
            $entityKeyValue = $row[$entityKeyName];
            if (is_null($entityKeyValue)) {
                throw new \Exception('No value for one of the entity keys was informed.');
            }

            $findData[strtolower($entityKeyName)] = $entityKeyValue;
        }

        return $findData;
    }

    private function buildSetEntityMethod(string $field): string {
        return 'set' . ucwords(strtolower($field));
    }
    
    private function getNewEntityDefaultFields(): array {
        $date = new \DateTime();
        
        $defaultFields = [
            'NRORG'             => Utils::getParamters()->getNrorg(),
            'IDATIVO'           => 'S',
            'NRORGINCLUSAO'     => Utils::getParamters()->getNrorg(),
            'DTINCLUSAO'        => $date,
            'CDOPERINCLUSAO'    => Utils::getParamters()->getCdoperador(),
            'NRORGULTATU'       => Utils::getParamters()->getNrorg(),
            'DTULTATU'          => $date,
            'CDOPERULTATU'      => Utils::getParamters()->getCdoperador()
        ];
        
        return $defaultFields;
    }

    private function getUpdatedEntityDefaultFields(): array {
        $date = new \DateTime();
        
        $defaultFields = [
            'NRORGULTATU'       => Utils::getParamters()->getNrorg(),
            'DTULTATU'          => $date,
            'CDOPERULTATU'      => Utils::getParamters()->getCdoperador()
        ];
        
        return $defaultFields;
    }

    private function unsetEntityKeysOnRow(
        array   $row,
        array   $entityKeysNames
    ): array {
        foreach ($entityKeysNames as $keyName) {
            unset($row[$keyName]);
        }

        return $row;
    }

    private function getEntityManager(): EntityManagerInterface {
        return Utils::getEntityManager();
    }
    
} 